# ===============================================
# (C) 2018 AO Kaspersky Lab. All Rights Reserved.
# ===============================================

import sys
import csv
import re
import binascii
import sqlite3
import hashlib
import socket

def CheckDate(s):
  if s==None or s=='':
    return None
  else:
    if re.match("^(19)|(20)|(21)\\d{2}-\\d{2}-\\d{2}$",s):
      return s
  return None

def Str2Date(s):
  if s == None or s=='':
    return None
  if not CheckDate(s):
    return None
  y = int(s[0:4])
  m = int(s[5:7])
  d = int(s[8:10])
  return (((y * 100) + m) * 100) + d

def Str2Date0(s):
  if s == None or s=='':
    return None
  return Str2Date(s)

def Str2DateT0(s):
  if s == None or s=='':
    return None
  return Str2Date(s[0:10])

def Date2Str(d):
  if d == None or d == 0:
    return None
  y,md = divmod(d,10000)
  m,d = divmod(md,100)
  return '%04d-%02d-%02d' % (y,m,d)

def DomainGood(s):
  if s==None or s=='' or s.find('/') != -1:
    return False
  return True

def Hex2Bin(s):
  pfx = s[0:2]
  if pfx=='0x' or pfx=='0X':
    s = s[2:]
  try:
    return binascii.a2b_hex(s)
  except Exception:
    return None

def Str2Bool(s):
  if s == None or s == '':
    return None
  s = s.lower()
  if s == 'true' or s == '1':
    return True
  if s == 'false' or s == '0':
   return False
  raise ValueError('Invalid bool value: %s' % (s,))

def RecodeBytes(s,key):
  kl = len(key)
  if kl == 0:
    return s
  sl = len(s)
  rs = bytearray()
  i = 0
  while i < sl:
    rs.append(s[i]^key[i%kl])
    i += 1
  return rs


def IpIs6(ip):
  if str(ip).find(':') == -1:
    return False
  else:
    return True

if hasattr(socket,'inet_pton') and hasattr(socket,'inet_ntop'):

  def Ip2Bin(ip):
    ip = str(ip)
    if IpIs6(ip):
      return socket.inet_pton(socket.AF_INET6,ip)
    else:
      return socket.inet_pton(socket.AF_INET,ip)

  def Bin2Ip(ipb):
    if len(ipb) != 4:
      return socket.inet_ntop(socket.AF_INET6,ipb)
    else:
      return socket.inet_ntop(socket.AF_INET,ipb)

else:

  def Ip2Bin(ip):
    ip = str(ip)
    if IpIs6(ip):
      raise Exception('No IP6 address support for this platform')
    else:
      return socket.inet_aton(ip)

  def Bin2Ip(ipb):
    if len(ipb) != 4:
      raise Exception('No IP6 address support for this platform')
    else:
      return socket.inet_ntoa(ipb)

def IpToDb(ip):
  ipb = Ip2Bin(ip)
  if len(ipb) == 4:
    ipb = bytes([0]*12) + ipb
  return ipb

csv.register_dialect('ipinfo', delimiter=',', quoting=csv.QUOTE_ALL)
csv.register_dialect('whois',  delimiter='\t', quoting=csv.QUOTE_NONE)

class SnapHostTop:
  def __init__(self,dbName,tableName):
    self.name = 'HostTop'
    self.db_name = dbName
    self.table_name = tableName

  def Prepare(self,conn):
    cursor = conn.cursor()
    cursor.execute('''CREATE TABLE IF NOT EXISTS %s (
        HK CHARACTER(16) PRIMARY KEY NOT NULL,
        AR INTEGER NULL,
        RD INTEGER NULL,
        RW INTEGER NULL,
        RM INTEGER NULL,
        UD INTEGER NULL,
        UW INTEGER NULL,
        UM INTEGER NULL
      )''' % (self.table_name,))

  def Import(self,fileName):
    conn = sqlite3.connect(self.db_name)
    self.Prepare(conn)

    print('Importing %s to %s:%s' % (fileName, self.db_name, self.table_name))
    cnt = 0
    firstSkipped = False

    cursor = conn.cursor()
    with open(fileName,encoding='utf-8-sig') as f:
      dialect = csv.Sniffer().sniff(f.read(1024))
      f.seek(0)
      reader = csv.reader(f,dialect)
      for row in reader:
        if not firstSkipped:
          firstSkipped = True
          continue
        name = row[0]
        if not DomainGood(name):
          print('skipped: %s' % (name,))
          continue

        i = 1
        while i < len(row):
          if row[i] == 'NULL':
            row[i] = None
          i += 1

        nameHash = hashlib.md5(name.encode('utf-8')).digest()

        query = "insert or ignore into " + self.table_name
        query += ' (HK,AR,RD,RW,RM,UD,UW,UM)'
        query += ' values (?,?,?,?,?,?,?,?)'
        try:
          cursor.execute(query, (nameHash,row[1],row[2],row[3],row[4],row[5],row[6],row[7]))
          cnt += 1
          if cnt % 10000 == 0:
            print('imported %s records' % (cnt,))
        except Exception as e:
          print('Exception: %s while executing query: %s' % (e, query))

    print('imported %s records' % (cnt,))
    print('Committing changes...')
    conn.commit()
    conn.close()

class SnapIpInfo:
  def __init__(self,dbName,tableName):
    self.name = 'IpInfo'
    self.db_name = dbName
    self.table_name = tableName

  def Prepare(self,conn):
    cursor = conn.cursor()
    cursor.execute('''CREATE TABLE IF NOT EXISTS %s (
        FromIp CHARACTER(16) PRIMARY KEY NOT NULL,
        ToIp   CHARACTER(16) NOT NULL,
        NetName NVARCHAR(500) NULL,
        OrgName NVARCHAR(500) NULL,
        Country NVARCHAR(500) NULL,
        City    NVARCHAR(500) NULL,
        Descr   NVARCHAR(500) NULL
      )''' % (self.table_name,))

  def Import(self,fileName):
    conn = sqlite3.connect(self.db_name)
    self.Prepare(conn)

    print('Importing %s to %s' % (fileName, self.table_name))
    cnt = 0

    cursor = conn.cursor()
    with open(fileName,encoding='utf-8-sig') as f:
      reader = csv.reader(f,'ipinfo')
      for row in reader:
        ipFromStr = row[0]
        ipToStr = row[1]
        if ipFromStr == '' or IpIs6(ipFromStr):
          continue

        ipFromBin = IpToDb(ipFromStr)
        ipToBin = IpToDb(ipToStr)

        netName = '%s-%s' % (ipFromStr,ipToStr)

        query = "insert or ignore into " + self.table_name
        query += ' (FromIp,ToIp,NetName,OrgName,Country,City,Descr)'
        query += ' values (?,?,?,?,?,?,?)'
        try:
          cursor.execute(query, (ipFromBin,ipToBin,netName,row[5],row[2],row[4],None))
          cnt += 1
          if cnt % 10000 == 0:
            print('imported %s records' % (cnt,))
        except Exception as e:
          print('Exception: %s while executing query: %s' % (e, query))

    print('imported %s records' % (cnt,))
    print('Committing changes...')
    conn.commit()
    conn.close()


class SnapWhois:
  def __init__(self,dbName,tableName):
    self.name = 'Whois'
    self.db_name = dbName
    self.table_name = tableName

  def Prepare(self,conn):
    cursor = conn.cursor()
    cursor.execute('''CREATE TABLE IF NOT EXISTS %s (
      HK CHAR(16) PRIMARY KEY NOT NULL,
      CD INT NULL,
      XD INT NULL,
      UD INT NULL,
      OD VARCHAR(500),
      RD VARCHAR(500),
      OG BOOLEAN NULL
    )''' % (self.table_name,))

  def Import(self,fileName):
    conn = sqlite3.connect(self.db_name)
    self.Prepare(conn)

    print('Importing %s to %s:%s' % (fileName, self.db_name, self.table_name))
    cnt = 0
    firstSkipped = False

    cursor = conn.cursor()
    with open(fileName,encoding='utf-8-sig') as f:
      reader = csv.reader(f,'whois')
      for row in reader:
        if not firstSkipped:
          firstSkipped = True
          continue
        domain = row[0]
        if not DomainGood(domain) or len(row)!=8:
          print('skipped: %s' % (domain,))
          continue

        domainKey  = hashlib.md5((domain*2).encode('utf-8')).digest()
        domainHash = hashlib.md5(domain.encode('utf-8')).digest()

        ownData = RecodeBytes(row[4].encode('utf-8'),domainHash)
        regData = RecodeBytes(row[6].encode('utf-8'),domainHash)
        cdate = Str2Date(row[1])
        xdate = Str2Date(row[5])
        udate = Str2Date(row[2])
        ownIsOrg = Str2Bool(row[7])

        query = "insert or ignore into " + self.table_name
        query += ' (HK,CD,XD,UD,OD,RD,OG)'
        query += ' values (?,?,?,?,?,?,?)'
        try:
          cursor.execute(query, (domainKey,cdate,xdate,udate,ownData,regData,ownIsOrg))
          cnt += 1
          if cnt % 10000 == 0:
            print('imported %s records' % (cnt,))
        except Exception as e:
          print('Exception: %s while executing query: %s' % (e, query))

    print('imported %s records' % (cnt,))
    print('Committing changes...')
    conn.commit()
    conn.close()


class SnapVhs:
  def __init__(self,dbName,tableName):
    self.name = 'VHS'
    self.db_name = dbName
    self.table_name = tableName

  def Prepare(self,conn):
    cursor = conn.cursor()
    cursor.execute('''CREATE TABLE IF NOT EXISTS %s (
      VH CHARACTER(16) PRIMARY KEY NOT NULL,
      TF INTEGER,
      TL INTEGER,
      CU INTEGER,
      CF INTEGER,
      CG INTEGER,
      CB INTEGER
    )''' % (self.table_name,))

  def Import(self,fileName):
    conn = sqlite3.connect(self.db_name)
    self.Prepare(conn)

    print('Importing %s to %s' % (fileName, self.table_name))
    cnt = 0

    cursor = conn.cursor()
    with open(fileName,encoding='utf-8-sig') as f:
      dialect = csv.Sniffer().sniff(f.read(1024))
      f.seek(0)
      reader = csv.reader(f, dialect)
      for row in reader:
        hashBin = Hex2Bin(row[0])
        if hashBin == None:
          print('skipped: %s' % (row[0],))
          continue

        query = "insert or ignore into " + self.table_name
        query += ' (VH,TF,TL,CU,CF,CG,CB)'
        query += ' values (?,?,?,?,?,?,?)'
        try:
          cursor.execute(query, (sqlite3.Binary(hashBin),Str2DateT0(row[1]),Str2DateT0(row[2]),row[3],row[4],row[5],row[6]))
          cnt += 1
          if cnt % 10000 == 0:
            print('imported %s records' % (cnt,))
        except Exception as e:
          print('Exception: %s while executing query: %s' % (e, query))

    print('imported %s records' % (cnt,))
    print('Committing changes...')
    conn.commit()
    conn.close()


def ImportHashes(conn, fileName, tableName):
  print('Importing %s to %s' % (fileName, tableName))
  cursor = conn.cursor()
  cnt = 0
  with open(fileName, 'rb') as f:
    for lineBin in f:
      line = lineBin.decode('utf-8-sig','ignore').strip()
      row = line.split('\t')
      hashStr = row[0]
      rank = row[2]
      name = None
      hashBin = Hex2Bin(hashStr)
      if hashBin == None:
        print('skipped: %s' % (hashStr,))
        continue

      try:
        query = "insert or ignore into " + tableName
        if name == None:
          query += ' (Hash, Rank)'
          query += ' values (?,?)'
          cursor.execute(query, (sqlite3.Binary(hashBin), rank))
        else:
          query += ' (Hash, Rank, Name)'
          query += ' values (?,?,?)'
          cursor.execute(query, (sqlite3.Binary(hashBin), rank, name))
        cnt += 1
        if cnt % 10000 == 0:
          print('imported %s records' % (cnt,))
          if cnt % 100000 == 0:
            conn.commit()
      except Exception as e:
        print('Exception: %s while executing query: %s' % (e, query))

  print('imported %s records' % (cnt,))
  print('Committing changes...')
  conn.commit()


class SnapHashContent:
  def __init__(self,dbName,tableName):
    self.name = 'HashContent'
    self.db_name = dbName
    self.table_name = tableName

  def Prepare(self,conn):
    cursor = conn.cursor()
    cursor.execute('''CREATE TABLE IF NOT EXISTS %s (
      Hash CHARACTER(16) PRIMARY KEY NOT NULL,
      Rank INTEGER NOT NULL
    )''' % (self.table_name,))

  def Import(self,fileName):
    conn = sqlite3.connect(self.db_name)
    self.Prepare(conn)
    ImportHashes(conn, fileName, self.table_name)
    conn.close()


class SnapHashName:
  def __init__(self,dbName,tableName):
    self.name = 'HashName'
    self.db_name = dbName
    self.table_name = tableName

  def Prepare(self,conn):
    cursor = conn.cursor()
    cursor.execute('''CREATE TABLE IF NOT EXISTS %s (
      Hash CHARACTER(16) PRIMARY KEY NOT NULL,
      Rank INTEGER NOT NULL
    )''' % (self.table_name,))

  def Import(self,fileName):
    conn = sqlite3.connect(self.db_name)
    self.Prepare(conn)
    ImportHashes(conn, fileName, self.table_name)
    conn.close()

class SnapHashPath:
  def __init__(self,dbName,tableName):
    self.name = 'HashPath'
    self.db_name = dbName
    self.table_name = tableName

  def Prepare(self,conn):
    cursor = conn.cursor()
    cursor.execute('''CREATE TABLE IF NOT EXISTS %s (
      Hash CHARACTER(16) PRIMARY KEY NOT NULL,
      Rank INTEGER NOT NULL
    )''' % (self.table_name,))

  def Import(self,fileName):
    conn = sqlite3.connect(self.db_name)
    self.Prepare(conn)
    ImportHashes(conn, fileName, self.table_name)
    conn.close()

class SnapHashPathMd5:
  def __init__(self,dbName,tableName):
    self.name = 'HashPathMd5'
    self.db_name = dbName
    self.table_name = tableName

  def Prepare(self,conn):
    cursor = conn.cursor()
    cursor.execute('''CREATE TABLE IF NOT EXISTS %s (
      Hash CHARACTER(16) PRIMARY KEY NOT NULL,
      Rank INTEGER NOT NULL
    )''' % (self.table_name,))

  def Import(self,fileName):
    conn = sqlite3.connect(self.db_name)
    self.Prepare(conn)
    ImportHashes(conn, fileName, self.table_name)
    conn.close()


SnapTypes = {}
SnapTypes['domain_top'] = SnapHostTop('snap01.db', 'DT')
SnapTypes['host_top']   = SnapHostTop('snap02.db', 'HT')
SnapTypes['whois']      = SnapWhois('snap03.db','DD')
SnapTypes['vhs']        = SnapVhs('snap04.db', 'VH')
SnapTypes['hashcont']   = SnapHashContent('snap05.db','HC')
SnapTypes['ip_info']    = SnapIpInfo('snap06.db','IpInfo')
SnapTypes['hashname']   = SnapHashName('snap07.db','FN')
SnapTypes['hashpath']   = SnapHashPath('snap08.db','FP')
SnapTypes['hashpathmd5'] = SnapHashPathMd5('snap09.db', 'FPC')



if __name__ == "__main__":

  if len(sys.argv) < 3:
    print('Not enough arguments, use import_snapshot <snaptype> <src>')
    exit(1)

  snapType = sys.argv[1]
  fileName = sys.argv[2]

  snap = SnapTypes[snapType]

  print('Importing snapshot: %s' % (snap.name,))
  print('Using database: %s, table: %s' % (snap.db_name,snap.table_name))

  snap.Import(fileName)

  print('Done')


